from test_server.server import *  # noqa pylint: disable=wildcard-import
from test_server.structure import *  # noqa pylint: disable=wildcard-import
from test_server.error import *  # noqa pylint: disable=wildcard-import
from test_server.version import TEST_SERVER_VERSION

__version__ = TEST_SERVER_VERSION
